unit TcpObjects; interface
uses Classes, Windows, SysUtils, Contnrs,
     IPHlpAPI, IPHelper;

type
  // for filtering by state
  TTcpStateFilter = ( tcpfNone,
                      tcpfClosed,    tcpfListening,   tcpfSyn_sent,
                      tcpfSyn_rcvd,  tcpfEstablished, tcpfFin_wait1,
                      tcpfFin_wait2, tcpfClose_wait,  tcpfClosing,
                      tcpfLast_ack,  tcpfTime_wait,   tcpfDelete_tcb);
  TTcpStateFilterSet = set of TTcpStateFilter;

type
  TTcpRowAfterCreate = procedure ( Sender: TObject) of object;

type
  TTcpRow  = class(TObject)
  private
    FTcpRow : TMibTCPRow;
    FID: string;
    FTimestamp: TDateTime;
    function GetLocalAddr: string;
    function GetLocalPort: string;
    function GetRemoteAddr: string;
    function GetRemotePort: string;
    function GetState: string;
    function GetTcpRow: TMibTCPRow;
    function GetLocalAddrDW: DWORD;
    function GetLocalPortDW: DWORD;
    function GetRemoteAddrDW: DWORD;
    function GetRemotePortDW: DWORD;
    function GetStateDW: DWORD;
    procedure SetStateDW(const Value: DWORD);
    function GetLocalPortNum: string;
    function GetRemotePortNum: string;
  public
    constructor Create( TcpRow : TMibTCPRow); virtual;
    destructor Destroy; override;

    property TcpRow : TMibTCPRow read GetTcpRow;
    property ID : string read FID;

    // string representation
    property LocalAddr  : string read GetLocalAddr;
    property LocalPort  : string read GetLocalPort;
    property LocalPortNum : string read GetLocalPortNum;
    property RemoteAddr : string read GetRemoteAddr;
    property RemotePort : string read GetRemotePort;
    property RemotePortNum : string read GetRemotePortNum;
    property State      : string read GetState;
    property Timestamp  : TDateTime read FTimestamp;

    function PortName( Port : DWord; Numeric : boolean) : string;

    // DW representation
    property LocalAddrDW  : DWORD read GetLocalAddrDW;
    property LocalPortDW  : DWORD read GetLocalPortDW;
    property RemoteAddrDW : DWORD read GetRemoteAddrDW;
    property RemotePortDW : DWORD read GetRemotePortDW;
    property StateDW      : DWORD read GetStateDW write SetStateDW;
    //
    function LogString : string;
    end;

  TTcpListBase = class( TObjectList)
  private
    function GetItem(Index: Integer): TTcpRow;
    procedure SetItem(Index: Integer; const Value: TTcpRow);
  protected
  public
    constructor Create;
    destructor Destroy; override;
    property Items[Index: Integer]: TTcpRow read GetItem write SetItem; default;
    function FindByID( ID : string) : TTcpRow;
    end;

  TTcpOnGetRow = procedure ( Sender: TObject; TcpRow : TTcpRow; NewRow : boolean) of object;

  TTcpList = class( TTcpListBase)
  private
    FStateFilter: TTcpStateFilterSet;
    FShowLoopback: boolean;
    FLogging: boolean;
    FLogStream: TStream;
    FOnGetRow: TTcpOnGetRow;
    procedure SetStateFilter(const Value: TTcpStateFilterSet);
    procedure SetShowLoopback(const Value: boolean);
    procedure SetLogging(const Value: boolean);
  protected
    procedure GetTcpRows( RList : TTcpListBase);
    procedure RemoveFiltered;
    function Filtered( TR : TMibTCPRow) : boolean;
  public
    constructor Create;
    destructor Destroy; override;
    property OnGetRow : TTcpOnGetRow read FOnGetRow write FOnGetRow;
    function Refresh : boolean; // true if there are changes
    // Filters
    property StateFilter : TTcpStateFilterSet read FStateFilter write SetStateFilter;
    property ShowLoopback : boolean read FShowLoopback write SetShowLoopback;
    // Logging
    property Logging : boolean read FLogging write SetLogging;
    property LogStream : TStream read FLogStream write FLogStream;
    end;

implementation

{ TTcpRow }

constructor TTcpRow.Create(TcpRow: TMibTCPRow);
begin
  inherited Create;
  FTcpRow := TcpRow;
  FTimestamp := Now;
  FID := IntToHex( FTcpRow.dwLocalAddr, 8) + IntToHex( FTcpRow.dwLocalPort, 8) +
         IntToHex( FTcpRow.dwRemoteAddr, 8) + IntToHex( FTcpRow.dwRemotePort, 8);
end;

destructor TTcpRow.Destroy;
begin

  inherited;
end;

function TTcpRow.GetLocalAddr: string;
begin Result := IpAddr2Str( FTcpRow.dwLocalAddr); end;

function TTcpRow.GetLocalAddrDW: DWORD;
begin Result := FTcpRow.dwLocalAddr; end;

function TTcpRow.GetLocalPort: string;
begin Result := Port2Svc( Port2Wrd( FTcpRow.dwLocalPort)); end;

function TTcpRow.GetLocalPortDW: DWORD;
begin Result := FTcpRow.dwLocalPort; end;

function TTcpRow.GetLocalPortNum: string;
begin Result := Format( '%4d', [Port2Wrd( FTcpRow.dwLocalPort)] ); end;

function TTcpRow.GetRemoteAddr: string;
begin Result := IpAddr2Str( FTcpRow.dwRemoteAddr); end;

function TTcpRow.GetRemoteAddrDW: DWORD;
begin Result := FTcpRow.dwRemoteAddr; end;

function TTcpRow.GetRemotePort: string;
begin Result := Port2Svc( Port2Wrd( FTcpRow.dwRemotePort)); end;

function TTcpRow.GetRemotePortDW: DWORD;
begin Result := FTcpRow.dwRemotePort; end;

function TTcpRow.GetRemotePortNum: string;
begin Result := Format( '%4d', [Port2Wrd( FTcpRow.dwRemotePort)] ); end;

function TTcpRow.GetState: string;
begin Result := TCPConnState[ FTcpRow.dwState]; end;

function TTcpRow.GetStateDW: DWORD;
begin Result := FTcpRow.dwState; end;

function TTcpRow.GetTcpRow: TMibTCPRow;
begin Result := FTcpRow; end;

function TTcpRow.LogString: string;
begin
  Result := TimeToStr( FTimestamp)        + ' | ' +
            LocalAddr  + ':' + PortName( LocalPortDW, true)  + ' | ' +
            RemoteAddr + ':' + PortName( RemotePortDW, true) + ' | ' +
            State;

end;

function TTcpRow.PortName(Port: DWord; Numeric: boolean): string;
begin
  if   Numeric then Result := Format( '%5d', [Port2Wrd(Port)])
  else begin
    Result := Port2Svc( Port2Wrd( Port));
    Result := StringOfChar( ' ', 7 - Length( Result)) + Result; 
    end;
end;

procedure TTcpRow.SetStateDW(const Value: DWORD);
begin FTcpRow.dwState := Value; end;

{ TTcpList }

constructor TTcpListBase.Create;
begin
  inherited Create( true);

end;

destructor TTcpListBase.Destroy;
begin

  inherited;
end;

function TTcpListBase.FindByID(ID: string): TTcpRow;
var
  Cou : integer;
  Found : boolean;
begin
  Result := NIL;
  Cou := 0;
  Found := false;
  while (Cou < Count) AND ( NOT Found) do begin
    Found := Items[ Cou].ID = ID;
    if Found then begin
      Result := Items[ Cou];
      break;
      end;
    inc( Cou);  
    end;
end;

function TTcpListBase.GetItem(Index: Integer): TTcpRow;
begin Result := (inherited Items[Index]) as TTcpRow; end;

procedure TTcpListBase.SetItem(Index: Integer; const Value: TTcpRow);
begin inherited Items[Index] := Value as TObject; end;

procedure TTcpList.SetStateFilter(const Value: TTcpStateFilterSet);
begin
  if FStateFilter = Value then exit;
  FStateFilter := Value;
  RemoveFiltered;
end;

constructor TTcpList.Create;
begin
  inherited Create;
  FOnGetRow := NIL;
  FStateFilter := [];
  FShowLoopback := true;
  FLogStream := NIL;
  FLogging := false;
end;

destructor TTcpList.Destroy;
begin

  inherited;
end;

function TTcpList.Filtered(TR: TMibTCPRow): boolean;
begin
  // By State
  Result := TTcpStateFilter( TR.dwState) IN FStateFilter;
  if Result then exit;
  // by loopback
  if NOT FShowLoopback then Result := TR.dwLocalAddr = $0100007F;
  if Result then exit;
end;

procedure TTcpList.SetShowLoopback(const Value: boolean);
begin
  if FShowLoopback = Value then exit;
  FShowLoopback := Value;
  RemoveFiltered;
end;

procedure TTcpList.RemoveFiltered;
var
  TmpDelTR : TObjectList;
  Cou : integer;
begin
  // removing filtered connections
  TmpDelTR := TObjectList.Create( false);
  for Cou := 0 to Count - 1 do begin // marking deleted objects
    if NOT Filtered( Items[ Cou].TcpRow) then begin
      TmpDelTR.Add( Items[ Cou]);
      end;
    end;
  for Cou := 0 to TmpDelTR.Count - 1 do Remove( TmpDelTR.Items[ Cou]);
  TmpDelTR.Free;
end;

procedure TTcpList.SetLogging(const Value: boolean);
begin
  if FLogging = Value then exit;
  FLogging := Value;
  if FLogging then;
end;

{ TTcpList }

procedure TTcpList.GetTcpRows(RList: TTcpListBase);
var
  TCPRow        : TMIBTCPRow;
  i, NumEntries  : integer;
  TableSize     : DWORD;
  ErrorCode     : DWORD;
  pBuf          : PChar;
  TR            : TTcpRow;
begin
  if not Assigned( RList ) then EXIT;
  RList.Clear;
  // first call : get size of table
  TableSize := 0;
  NumEntries := 0 ;
  ErrorCode := GetTCPTable(Nil, @TableSize, false );
  if Errorcode <> ERROR_INSUFFICIENT_BUFFER then  EXIT;

  // get required memory size, call again
  GetMem( pBuf, TableSize );
  try
    // get table
    ErrorCode := GetTCPTable( PTMIBTCPTable( pBuf ), @TableSize, false );
    if ErrorCode <> NO_ERROR then Abort;

    NumEntries := PTMIBTCPTable( pBuf )^.dwNumEntries;
    if NumEntries <= 0 then Abort;

    inc( pBuf, SizeOf( DWORD ) ); // get past table size
    for i := 1 to NumEntries do begin
      TCPRow := PTMIBTCPRow( pBuf )^; // get next record
      with TCPRow do begin
        if dwRemoteAddr = 0 then dwRemotePort := 0;
        if NOT Filtered( TCPRow) then begin
          TR := TTcpRow.Create( TCPRow);
          RList.Add( TR);
          end;
        end;
      inc( pBuf, SizeOf( TMIBTCPRow ) );
      end;
  finally
    dec( pBuf, SizeOf( DWORD ) + NumEntries * SizeOf( TMibTCPRow ) );
    FreeMem( pBuf );
    end;
end;

function TTcpList.Refresh : boolean;
var
  TmpOL : TTcpListBase;
  TmpDelTR : TObjectList;
  Cou : integer;
  TmpTR, TR : TTcpRow;

  procedure WriteLog( TR : TTcpRow);
  var S : string;
  begin
    if Logging AND ( LogStream <> NIL) then begin // write log
      S := TR.LogString + #$0D#$0A;
      FLogStream.WriteBuffer( S[1], length( S));
      end;
  end;

begin
  Result := false;
  TmpOL := TTcpListBase.Create;

  // get new connections
  try GetTcpRows( TmpOL);
  except TmpOL.Free; raise; end;

  // removing terminated connections
  TmpDelTR := TObjectList.Create( false);
  for Cou := 0 to Count - 1 do begin // marking deleted objects
    if TmpOL.FindByID( Items[ Cou].ID) = NIL then begin
      TmpDelTR.Add( Items[ Cou]);
      Result := true;
      end;
    end;
  for Cou := 0 to TmpDelTR.Count - 1 do Remove( TmpDelTR.Items[ Cou]);
  TmpDelTR.Free;

  // edit existing connections or add new
  for Cou := 0 to TmpOL.Count - 1 do begin
    TmpTR := TmpOL.Items[ Cou];
    TR := FindByID( TmpTR.ID);
    if TR <> NIL then begin
      if TR.StateDW <> TmpTR.StateDW then begin
        TR.StateDW := TmpTR.StateDW; // change state
        if Assigned( FOnGetRow) then FOnGetRow( Self, TmpTR, false);
        WriteLog( TmpTR);
        Result := true;
        end;
      end
    else begin
      Add( TmpTR);
      if Assigned( FOnGetRow) then FOnGetRow( Self, TmpTR, true);
      WriteLog( TmpTR);
      Result := true;
      end;
    end;
end;

end.
